#define ALLEGRO_UNSTABLE
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <allegro5/allegro.h>
#include <allegro5/allegro_image.h>
#include <allegro5/allegro_font.h>
#include <allegro5/allegro_opengl.h>
#include <allegro5/allegro_native_dialog.h>

#define BUFFER_WIDTH  640
#define BUFFER_HEIGHT 480

#define WINDOW_WIDTH  640
#define WINDOW_HEIGHT 480

ALLEGRO_DISPLAY *display = NULL;
ALLEGRO_BITMAP *buffer;
ALLEGRO_FONT *font;

volatile bool window_closed = false;

void abort_on_error(const char *message) {
	fprintf(stderr, "%s\n", message);
	al_show_native_message_box(display, "Error", "An error has occurred", message, NULL, ALLEGRO_MESSAGEBOX_ERROR);
	exit(EXIT_FAILURE);
}

void setup() {

	if (!al_install_system(ALLEGRO_VERSION_INT, NULL))
		abort_on_error("Failed to init Allegro");
	if (!al_install_keyboard())
		abort_on_error("Failed to init Allegro keyboard drivers");
	if (!al_init_image_addon())
		abort_on_error("Failed to init Allegro image add-on");
	al_init_font_addon();

	al_set_new_display_flags(ALLEGRO_OPENGL | ALLEGRO_WINDOWED);
	al_set_new_display_option(ALLEGRO_DEPTH_SIZE, 16, ALLEGRO_SUGGEST);
	
	display = al_create_display(WINDOW_WIDTH, WINDOW_HEIGHT);
	if (!display) abort_on_error("Failed to create window");
   
	al_set_window_title(display, "Allegro OpenGL Test");
	
	// Creates bitmap buffer
	al_set_new_bitmap_depth(16);
    buffer = al_create_bitmap(BUFFER_WIDTH, BUFFER_HEIGHT);
    if (!buffer) abort_on_error("Failed to create video buffer");
	al_set_new_bitmap_depth(0);
	
    { // Load bitmap font
		ALLEGRO_BITMAP *bmp_font = al_load_bitmap("font.png");
		if (!bmp_font) abort_on_error("Failed to load bitmap font");
		int ranges[] = {0x20,0x7e}, num_ranges=1;
		font = al_grab_font_from_bitmap(bmp_font, num_ranges, ranges);
		al_destroy_bitmap(bmp_font);
	}
}

void shutdown() {
	al_destroy_font(font);
    al_destroy_bitmap(buffer);
	al_destroy_display(display);
	al_uninstall_system();
}

// Draws a simple colored cube
void draw_cube() {
	glBegin(GL_POLYGON);
		glColor3f(0.0,0.0,1.0);
		glVertex3f( 1.0,-1.0, 1.0);
		glVertex3f(-1.0,-1.0, 1.0);
		glVertex3f(-1.0, 1.0, 1.0);
		glVertex3f( 1.0, 1.0, 1.0);
	glEnd();
	
	glBegin(GL_POLYGON);
		glColor3f(0.0,1.0,0.0);
		glVertex3f(-1.0,-1.0,-1.0);
		glVertex3f(-1.0, 1.0,-1.0);
		glVertex3f(-1.0, 1.0, 1.0);
		glVertex3f(-1.0,-1.0, 1.0);
	glEnd();

	glBegin(GL_POLYGON);
		glColor3f(0.0,1.0,1.0);
		glVertex3f(-1.0,-1.0, 1.0);
		glVertex3f( 1.0,-1.0, 1.0);
		glVertex3f( 1.0,-1.0,-1.0);
		glVertex3f(-1.0,-1.0,-1.0);
	glEnd();
	
	glBegin(GL_POLYGON);
		glColor3f(1.0,0.0,0.0);
		glVertex3f(-1.0,-1.0,-1.0);
		glVertex3f( 1.0,-1.0,-1.0);
		glVertex3f( 1.0, 1.0,-1.0);
		glVertex3f(-1.0, 1.0,-1.0);
	glEnd();
	
	glBegin(GL_POLYGON);
		glColor3f(1.0,0.0,1.0);
		glVertex3f( 1.0, 1.0,-1.0);
		glVertex3f( 1.0,-1.0,-1.0);
		glVertex3f( 1.0,-1.0, 1.0);
		glVertex3f( 1.0, 1.0, 1.0);
	glEnd();
	
	glBegin(GL_POLYGON);
		glColor3f(1.0,1.0,0.0);
		glVertex3f( 1.0, 1.0, 1.0);
		glVertex3f(-1.0, 1.0, 1.0);
		glVertex3f(-1.0, 1.0,-1.0);
		glVertex3f( 1.0, 1.0,-1.0);
	glEnd();
}

// Like al_draw_text, but adds a hard shadow to font
void draw_text_shadow (ALLEGRO_FONT *font, ALLEGRO_COLOR color_text,
    ALLEGRO_COLOR color_shadow, int x, int y, int flags, const char *text) {
    al_draw_text(font, color_shadow, x+1, y+1, flags, text);
    al_draw_text(font, color_text  , x  , y  , flags, text);
}

// Flushes buffer to screen
void flush_buffer() {
    al_set_target_backbuffer(display);
    al_draw_scaled_bitmap(buffer, 0, 0, BUFFER_WIDTH, BUFFER_HEIGHT,
        0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, 0);
    al_flip_display();
    al_set_target_bitmap(buffer);
}

// Equivalent to gluPerspective
void set_perpective(float angle, float aspect, float z_near, float z_far) {
	float fH = tan((angle/360.0f*M_PI));
	float fW = fH * aspect;
	glFrustum(-fW, fW, -fH, fH, z_near, z_far);
}

void play() {
	ALLEGRO_KEYBOARD_STATE k_state;
	ALLEGRO_COLOR
		white = al_map_rgb_f(1,1,1),
		gray = al_map_rgb_f(.5,.5,.5);
	
	ALLEGRO_BITMAP *disc = al_load_bitmap("disc.png");
	if (!disc) abort_on_error("Failed to open image");
	float cx = al_get_bitmap_width(disc)*.5,
	      cy = al_get_bitmap_height(disc)*.5;
	
	bool use_buffer=0, change_mode=1;
	int s_width, s_height;
	int i=0;
	float x_angle = 0, y_angle = 20, distance=-5;
	
	while (true) {
		i++;
		
		al_get_keyboard_state(&k_state);
		if (al_key_down(&k_state, ALLEGRO_KEY_ESCAPE)) break;
		
		if (al_key_down(&k_state, ALLEGRO_KEY_UP))   y_angle--;
		if (al_key_down(&k_state, ALLEGRO_KEY_DOWN)) y_angle++;
		
		if (al_key_down(&k_state, ALLEGRO_KEY_LEFT))  x_angle--;
		if (al_key_down(&k_state, ALLEGRO_KEY_RIGHT)) x_angle++;
		
		if (al_key_down(&k_state, ALLEGRO_KEY_PGUP)) distance-=.2;
		if (al_key_down(&k_state, ALLEGRO_KEY_PGDN)) distance+=.2;
		if (distance > 0) distance = 0;
		
		if (al_key_down(&k_state, ALLEGRO_KEY_SPACE)) {
			if (change_mode) use_buffer = !use_buffer;
			change_mode = false;
		}
		else
			change_mode = true;
		
		if (use_buffer) {
			s_width = BUFFER_WIDTH;
			s_height = BUFFER_HEIGHT;
		}
		else {
			s_width = WINDOW_WIDTH;
			s_height = WINDOW_HEIGHT;
		}
		
		// Prints text before OpenGL renders
		if (use_buffer) al_set_target_bitmap(buffer);
		else            al_set_target_backbuffer(display);
		glDisable(GL_DEPTH_TEST); // Need to disable depth test for Allegro drawing routines
		
		al_clear_to_color(al_map_rgb(0,0,64));
		
		draw_text_shadow(font, white, gray, 10, 10, 0, "allegro opengl test");
		draw_text_shadow(font, white, gray, 10, 30, 0, "[SPACE] toggle bitmap buffer mode");
		draw_text_shadow(font, white, gray, 10, 45, 0, "[ARROW KEYS] rotate view | [PGUP/PGDN] zoom");
		draw_text_shadow(font, white, gray, 10, 60, 0, "press [ESC] to exit");
		draw_text_shadow(font, white, gray, s_width-10, 10, ALLEGRO_ALIGN_RIGHT,
			use_buffer?"bitmap buffer":"display backbuffer");
		
		al_draw_rotated_bitmap(disc, cx, cy, s_width*.25, s_height*.5, i/60.0, 0);
				
		// OpenGL rendering
		
		// Set camera position
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		set_perpective(45, (float) s_width/s_height, 1, 1000);
		glTranslated(0, 0, distance);
		glRotatef(y_angle, 1.0, 0.0, 0.0);
		glRotatef(x_angle, 0.0, 1.0, 0.0);
		
		// Draw objects
		glMatrixMode(GL_MODELVIEW);
		glEnable(GL_DEPTH_TEST);
//		glDepthFunc(GL_LEQUAL);
//		glEnable(GL_CULL_FACE);
//		glFrontFace(GL_CW);
//		glEnable(GL_ALPHA_TEST);
		glClear(GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();
		
		glPushMatrix();
		glRotatef((float) i, 0.0, 1.0, 0.0);
		
		draw_cube();
		
		glPopMatrix();
		
		glPushMatrix();
		glTranslated(0, 0, -5);
		glRotatef((float) -i, 0.0, 1.0, 0.0);
		
		draw_cube();
		
		glPopMatrix();
		
		// Prints text over the OpenGL result
		if (use_buffer) al_set_target_bitmap(buffer);
		else            al_set_target_backbuffer(display);
		glDisable(GL_DEPTH_TEST);
		
		al_draw_rotated_bitmap(disc, cx, cy, s_width*.75, s_height*.5, i/60.0, 0);
		
		draw_text_shadow(font, white, gray, s_width/2, s_height*.75 - 5,
			ALLEGRO_ALIGN_CENTER, "text over");
		draw_text_shadow(font, white, gray, s_width/2, s_height*.75 + 5,
			ALLEGRO_ALIGN_CENTER, "rotating cube");
		
		// Flush buffer / page flip
		if (use_buffer) flush_buffer();
		else            al_flip_display();
	}
	
	al_destroy_bitmap(disc);
}

int main (int argc, char *argv[]) {
	setup();
	play();
	shutdown();
	return 0;
}
